/**
  *  \file planetscentral/host/cmdhost.cc
  *  \brief Host: HOST/MASTER commands
  *
  *  There's a parallel set of HOST and MASTER commands to manage host
  *  and master programs, respectively. They are implemented
  *  identically, their domain is handed in as a database subtree.
  */

#include <sys/stat.h>
#include <unistd.h>
#include "planetscentral/host/cmdhost.h"
#include "int/args.h"
#include "int/array.h"
#include "planetscentral/dbclient/dbhash.h"
#include "planetscentral/dbclient/dbset.h"
#include "planetscentral/dbclient/reconnect.h"
#include "planetscentral/host/host.h"
#include "planetscentral/osutil.h"
#include "planetscentral/host/rating.h"

/** Validate a tool identifier. Since these are used in shell variable
    names and URLs, they'd better be sane. We allow alphanumerics,
    ".-+_", which should be safe. Although they only come from
    controlled sources (admin installing new components), catching
    errors is a good idea. */
static bool
isValidIdentifier(const string_t& id, bool shellVar)
{
    if (id.empty()) {
        return false;
    }
    for (string_t::size_type i = 0; i < id.size(); ++i) {
        const char ch = id[i];
        if (!((ch >= 'A' && ch <= 'Z')
              || (ch >= 'a' && ch <= 'z')
              || (ch >= '0' && ch <= '9')
              || (ch == '_')
              || (!shellVar && (ch == '-' || ch == '.' || ch == '+' ))))
        {
            return false;
        }
    }
    return true;
}

IntValue*
describeHost(const DbSubtree& base, string_t name, bool isDefault)
{
    /* @type HostToolInfo
       Description of a host/master/shiplist/tool.
       @key id:Str           (tool Id)
       @key description:Str  (user-friendly description)
       @key kind:Str         (tool kind)
       @key default:Int      (1 if this tool is the default) */
    Ptr<IntArrayData> a = new IntArrayData();
    DbHash h(base.hashKey("prog:" + name));

    a->content.pushNew(makeStringValue("id"));
    a->content.pushNew(makeStringValue(name));
    a->content.pushNew(makeStringValue("description"));
    a->content.pushNew(makeStringValue(h.getStringField("description")));
    a->content.pushNew(makeStringValue("kind"));
    a->content.pushNew(makeStringValue(h.getStringField("kind")));
    a->content.pushNew(makeStringValue("default"));
    a->content.pushNew(makeIntValue(isDefault));
    a->addDimension(a->content.size());
    return new IntArray(a);
}

/** Validate executable. c2file cannot make programs executable. One
    solution would have been to add a CHMOD or PUTEXEC command.
    However, since c2file uses the PCC2 file abstraction, the 'chmod'
    call is not easily accessible.

    Since we ought to validate the program anyway, we can as well set
    the x bit ourselves.

    FIXME: this might need reconsideration when we implement proper
    live backup in c2file, which would have to back up the x bits as
    well. */
static bool
validateExecutable(string_t name)
{
    struct stat st;
    if (stat(name.c_str(), &st) != 0 || !S_ISREG(st.st_mode)) {
        // Not a file
        return false;
    }
    if (chmod(name.c_str(), st.st_mode | 0100) != 0) {
        // Not chmodable
        return false;
    }
    return true;
}

/* @q HOSTADD id:Str path:FileName program:Str kind:Str (Host Command)
   Add host.

   Permissions: admin.

   @err 412 Invalid identifier
   @err 412 Invalid kind
   @err 412 Invalid executable

   @uses prog:host:prog:$HOST, prog:host:list */

/* @q MASTERADD id:Str path:FileName program:Str kind:Str (Host Command)
   Add master.

   Permissions: admin.

   @err 412 Invalid identifier
   @err 412 Invalid kind
   @err 412 Invalid executable

   @uses prog:master:prog:$MASTER, prog:master:list */

/* @q TOOLADD id:Str path:FileName program:Str kind:Str (Host Command)
   Add tool/add-on.
   %program can be the empty string if the tool does not have a primary program.

   Permissions: admin.

   @err 412 Invalid identifier
   @err 412 Invalid kind
   @err 412 Invalid executable

   @uses prog:tool:prog:$TOOL, prog:tool:list */

/* @q SHIPLISTADD id:Str path:FileName program:Str kind:Str (Host Command)
   Add shiplist.

   Permissions: admin.

   @err 412 Invalid identifier
   @err 412 Invalid kind
   @err 412 Invalid executable

   %program should be the empty string as shiplists do not have a program.
   @uses prog:sl:prog:$SHIPLIST, prog:sl:list */
void
doHostAdd(const DbSubtree& base, HostConnection& conn, IntArgBlock& args)
{
    // HOSTADD <id> <path> <program> <kind>
    args.checkArgumentCount(4);
    conn.checkAdmin();

    ReconnectingGuard g(database_rc);
    string_t id   = toStringResult(args.getNext());
    string_t path = toStringResult(args.getNext());
    string_t prog = toStringResult(args.getNext());
    string_t kind = toStringResult(args.getNext());

    if (!isValidIdentifier(id, false)) {
        conn.sendError("412 Invalid identifier");
    } else if (!isValidIdentifier(kind, true)) {
        conn.sendError("412 Invalid kind");
    } else if (!prog.empty() && !validateExecutable(path + "/" + prog)) {
        conn.sendError("412 Invalid executable");
    } else {
        DbHash h(base.hashKey("prog:" + id));
        h.setField("path", path);
        h.setField("program", prog);
        h.setField("kind", kind);
        base.stringSetKey("list").add(id);
        if (base.stringKey("default").get().empty()) {
            base.stringKey("default").set(id);
        }
        conn.sendSuccess("OK");
    }
}

/* @q HOSTSET id:Str key:Str value:Str (Host Command), MASTERSET id:Str key:Str value:Str (Host Command), SHIPLISTSET id:Str key:Str value:Str (Host Command), TOOLSET id:Str key:Str value:Str (Host Command)
   Set host/master/shiplist/tool property.

   Permissions: admin.

   @err 412 Invalid identifier

   @uses prog:host:prog:$HOST, prog:tool:prog:$TOOL, prog:sl:prog:$SHIPLIST, prog:master:prog:$MASTER */
void
doHostSet(const DbSubtree& base, HostConnection& conn, IntArgBlock& args)
{
    // HOSTSET <id> <key> <value>
    args.checkArgumentCount(3);
    conn.checkAdmin();

    // FIXME: validate that <id> exists? validate <key>?
    string_t id  = toStringResult(args.getNext());
    string_t key = toStringResult(args.getNext());
    string_t val = toStringResult(args.getNext());

    if (!isValidIdentifier(id, false)) {
        conn.sendError("412 Invalid identifier");
    } else {
        ReconnectingGuard g(database_rc);
        base.hashKey("prog:" + id).setField(key, val);
        conn.sendSuccess("OK");
    }
}

/* @q HOSTGET id:Str key:Str (Host Command), MASTERGET id:Str key:Str (Host Command), SHIPLISTGET id:Str key:Str (Host Command), TOOLGET id:Str key:Str (Host Command)
   Get host/master/shiplist/tool property.

   Permissions: none.

   @retval Str property value
   @uses prog:host:prog:$HOST, prog:tool:prog:$TOOL, prog:sl:prog:$SHIPLIST, prog:master:prog:$MASTER */
void
doHostGet(const DbSubtree& base, HostConnection& conn, IntArgBlock& args)
{
    // HOSTGET <id> <key>
    args.checkArgumentCount(2);

    // FIXME: validate that <id> exists? validate <key>?
    // Not critical; although users can fetch arbitrary keys using
    // "toolinfo.cgi/host/some-random-string", they cannot do anything
    // evil with it.
    string_t id  = toStringResult(args.getNext());
    string_t key = toStringResult(args.getNext());

    ReconnectingGuard g(database_rc);
    IntStringValue sv(base.hashKey("prog:" + id).getStringField(key));
    conn.sendObject(&sv);
}

/* @q HOSTRM id:Str (Host Command), MASTERRM id:Str (Host Command), SHIPLISTRM id:Str (Host Command), TOOLRM id:Str (Host Command)
   Remove a host/master/shiplist/tool.

   Permissions: admin.

   @uses prog:host:prog:$HOST, prog:tool:prog:$TOOL, prog:sl:prog:$SHIPLIST, prog:master:prog:$MASTER
   @uses prog:host:list, prog:tool:list, prog:sl:list, prog:master:list
   @retval Int (1=removed, 0=did not exist) */
void
doHostRemove(const DbSubtree& base, HostConnection& conn, IntArgBlock& args)
{
    // HOSTRM <name>
    args.checkArgumentCount(1);
    conn.checkAdmin();

    string_t name = toStringResult(args.getNext());
    ReconnectingGuard g(database_rc);

    bool result = base.stringSetKey("list").remove(name);
    if (result) {
        base.hashKey("prog:" + name).remove();
    }

    if (base.stringKey("default").get() == name) {
        // This was the default, pick another one
        base.stringKey("default").set(base.stringSetKey("list").getRandom());
    }

    IntIntBoolValue iv(result);
    conn.sendObject(&iv);
}

/* @q HOSTLS (Host Command), MASTERLS (Host Command), SHIPLISTLS (Host Command), TOOLLS (Host Command)
   List hosts/masters/shiplists/tools.

   Permissions: none.

   @retval HostToolInfo[] list of all requested components
   @uses prog:host:prog:$HOST, prog:tool:prog:$TOOL, prog:sl:prog:$SHIPLIST, prog:master:prog:$MASTER
   @uses prog:host:list, prog:tool:list, prog:sl:list, prog:master:list */
void
doHostList(const DbSubtree& base, HostConnection& conn, IntArgBlock& /*args*/)
{
    // HOSTLS
    ReconnectingGuard g(database_rc);

    string_t def = base.stringKey("default").get();

#if 1
    /* Simple regular way. */
    StringList_t hostList;
    base.stringSetKey("list").getAll(hostList);
    Ptr<IntArrayData> result = new IntArrayData;
    for (uint32_t i = 0, n = hostList.size(); i < n; ++i) {
        result->content.pushNew(describeHost(base, hostList[i], def==hostList[i]));
    }
#else
    /* Ugly special-purpose way. This uses SORT to obtain the complete result
       in one redis request. It is not faster than the simple regular way.
       It is kept here for reference. */
    std::auto_ptr<IntValue> value(database_connection.call(DbRequest()
                                                           .withString("SORT")
                                                           .withString(base.getPrefix() + "list")
                                                           .withString("BY").withString("nosort")
                                                           .withString("GET").withString("#")
                                                           .withString("GET").withString(base.getPrefix() + "prog:*->description")
                                                           .withString("GET").withString(base.getPrefix() + "prog:*->name")));
    StringList_t list;
    toStringListResult(value.get(), list);
    Ptr<IntArrayData> result = new IntArrayData;
    for (uint32_t i = 0; i+3 <= list.size(); i += 3) {
        // Result will be hash of id, description, kind
        Ptr<IntArrayData> aa = new IntArrayData();
        aa->content.pushNew(makeStringValue("id"));
        aa->content.pushNew(makeStringValue(list[i]));
        aa->content.pushNew(makeStringValue("description"));
        aa->content.pushNew(makeStringValue(list[i+1]));
        aa->content.pushNew(makeStringValue("kind"));
        aa->content.pushNew(makeStringValue(list[i+2]));
        aa->content.pushNew(makeStringValue("default"));
        aa->content.pushNew(makeIntValue(def == list[i]));
        aa->addDimension(aa->content.size());
        result->content.pushNew(new IntArray(aa));
    }
#endif
    result->addDimension(result->content.size());
    IntArray a(result);
    conn.sendObject(&a);
}

/* @q HOSTCP src:Str dest:Str (Host Command), MASTERCP src:Str dest:Str (Host Command), SHIPLISTCP src:Str dest:Str (Host Command), TOOLCP src:Str dest:Str (Host Command)
   Copy host/master/shiplist/tool.

   Permissions: admin.

   @err 404 Missing source (invalid %src)
   @err 412 Invalid identifier (invalid %dest)

   @uses prog:host:prog:$HOST, prog:tool:prog:$TOOL, prog:sl:prog:$SHIPLIST, prog:master:prog:$MASTER
   @uses prog:host:list, prog:tool:list, prog:sl:list, prog:master:list */
void
doHostCopy(const DbSubtree& base, HostConnection& conn, IntArgBlock& args)
{
    // HOSTCP <source> <dest>
    args.checkArgumentCount(2);
    conn.checkAdmin();

    string_t from = toStringResult(args.getNext());
    string_t to   = toStringResult(args.getNext());

    ReconnectingGuard g(database_rc);
    DbStringSet list(base.stringSetKey("list"));

    // Validate source
    if (!list.contains(from)) {
        conn.sendError("404 Missing source");
        return;
    }
    if (!isValidIdentifier(to, false)) {
        conn.sendError("412 Invalid identifier");
        return;
    }

    // Create copy
    StringList_t data;
    base.hashKey("prog:" + from).getAll(data);
    base.hashKey("prog:" + to).remove();
    base.hashKey("prog:" + to).setAll(data);

    list.add(to);
    conn.sendSuccess("OK");
}

/* @q HOSTDEFAULT id:Str (Host Command), MASTERDEFAULT id:Str (Host Command), SHIPLISTDEFAULT id:Str (Host Command)
   Set default host/master/shiplist.

   Permissions: admin.

   @err 404 Not found (%id does not exist)

   @uses prog:host:prog:$HOST, prog:tool:prog:$TOOL, prog:sl:prog:$SHIPLIST, prog:master:prog:$MASTER
   @uses prog:host:list, prog:tool:list, prog:sl:list, prog:master:list */
void
doHostDefault(const DbSubtree& base, HostConnection& conn, IntArgBlock& args)
{
    // HOSTDEFAULT <id>
    args.checkArgumentCount(1);
    conn.checkAdmin();

    string_t what = toStringResult(args.getNext());

    ReconnectingGuard g(database_rc);

    // Validate
    if (!base.stringSetKey("list").contains(what)) {
        conn.sendError("404 Not found");
        return;
    }

    // Set default
    base.stringKey("default").set(what);
    conn.sendSuccess("OK");
}

/* @q HOSTRATING id:Str {{SET n:Int | AUTO} {USE|SHOW}}|NONE|GET (Host Command), MASTERRATING id:Str {{SET n:Int | AUTO} {USE|SHOW}}|NONE|GET (Host Command), SHIPLISTRATING id:Str {{SET n:Int | AUTO} {USE|SHOW}}|NONE|GET (Host Command), TOOLRATING id:Str {{SET n:Int | AUTO} {USE|SHOW}}|NONE|GET (Host Command)
   Configure host/master/shiplist/tool rating.
   - "SET n" sets the rating to the given value.
   - "AUTO" resets the rating to the computed default value.<br/>
     SET or AUTO followed by "USE" actually uses the specified difficulty,
     followed by "SHOW" only shows it but internally uses the computed one.
   - "NONE" resets the difficulty.
   - "GET" does not modify the difficulty; instead, returns it.

   Permissions: admin.

   @err 404 Not found
   @err 400 Syntax error

   @retval Int new difficulty (unless NONE was used)
   @uses prog:host:prog:$HOST, prog:tool:prog:$TOOL, prog:sl:prog:$SHIPLIST, prog:master:prog:$MASTER */
void
doHostRating(const DbSubtree& base, HostConnection& conn, IntArgBlock& args)
{
    // HOSTRATING <id> {{SET n|AUTO} {USE|SHOW} | NONE}
    args.checkArgumentCountAtLeast(2);
    conn.checkAdmin();

    // Parse
    string_t what = toStringResult(args.getNext());
    string_t op = strUCase(toStringResult(args.getNext()));
    enum { Set, Auto, None, Get } mode;
    int32_t value = 0;
    int32_t use = 0;

    // First verb
    if (op == "SET") {
        args.checkArgumentCount(2);
        mode = Set;
        value = toIntResult(args.getNext());
    } else if (op == "AUTO") {
        args.checkArgumentCount(1);
        mode = Auto;
    } else if (op == "NONE") {
        args.checkArgumentCount(0);
        mode = None;
    } else if (op == "GET") {
        args.checkArgumentCount(0);
        mode = Get;
    } else {
        conn.sendError("400 Syntax error");
        return;
    }

    // Modifier for SET/AUTO
    if (mode == Set || mode == Auto) {
        string_t useWord =  strUCase(toStringResult(args.getNext()));
        if (useWord == "USE") {
            use = 1;
        } else if (useWord == "SHOW") {
            use = 0;
        } else {
            conn.sendError("400 Syntax error");
            return;
        }
    }

    // Operate
    ReconnectingGuard g(database_rc);
    if (!base.stringSetKey("list").contains(what)) {
        conn.sendError("404 Not found");
        return;
    }

    DbHash tool(base.subtree("prog").hashKey(what));
    if (mode == Auto) {
        value = computeToolRating(tool.getStringField("path"));
    }
    if (value < 1) {
        value = 1;
    }
    if (value > 1000) {
        value = 1000;
    }

    if (mode == Get) {
        value = tool.hasField("difficulty") ? tool.getIntField("difficulty") : 0;
    } else if (mode == None) {
        tool.removeField("difficulty");
        tool.removeField("useDifficulty");
    } else {
        tool.setField("difficulty", value);
        tool.setField("useDifficulty", use);
    }

    if (mode == None) {
        conn.sendSuccess("OK");
    } else {
        IntIntIntValue iv(value);
        conn.sendObject(&iv);
    }
}
