/**
  *  \file planetscentral/mailout/transmitter.h
  *  \brief Mailout: Mail Transmitter
  */
#ifndef C2SERVER_PLANETSCENTRAL_MAILOUT_TRANSMITTER_H
#define C2SERVER_PLANETSCENTRAL_MAILOUT_TRANSMITTER_H

#include <list>
#include "cpluslib/types.h"
#include "cpluslib/string.h"
#include "planetscentral/osutil.h"

class Message;

/** Transmitter. Messages are always stored in the redis database for
    persistance, although they often are short-lived.

    There are two queues in redis:
    - MAIL_ROOT->sending
    - MAIL_ROOT->preparing

    Upon startup, we delete all messages from the preparing queue;
    these are messages that were partially prepared but not sent.

    The sending queue is mirrored in RAM.
    - work_queue
    - pend_queue

    All messages are placed in work_queue first. After they are sent,
    they are removed. If they cannot be sent right now, they are moved
    to the pend_queue and reconsidered at a later time by moving them
    back to work_queue. */
class Transmitter {
 public:
    Transmitter();

    void start();
    void setTemplateDir(string_t dir);
    void send(int32_t id);
    void notifyAddress(string_t adr);
    void runQueue();
    bool maybeRequestConfirmation(string_t user);

    static bool checkConfig(string_t key, string_t value);

 private:
    static void* entry(void*);
    void processWork();
    bool sendMessage(Message& msg, string_t address);
    bool resolveAddress(string_t address, string_t& smtpAddress, string_t& authUser);
    void requestConfirmation(string_t user, string_t userEmail);

    /* Directory containing template files */
    string_t template_dir;

    /* Mutex protecting work_queue and pend_queue. Lowest level of the locking
       hierarchy; you may not obtain any other mutex while holding this one. */
    Mutex queue_mutex;

    /* Work semaphore. Mirrors the number of items in work_queue. */
    Semaphore work_sem;

    /* Queue of items to process. Head is currently being worked on. */
    std::list<int32_t> work_queue;

    /* Queue of items that failed because of an unverified address. */
    std::list<int32_t> pend_queue;
};

int32_t getCurrentTime();
int32_t getMaximumMessageAge();
bool confirmMail(string_t mail, string_t key, string_t info);

#endif
