/**
  *  \file planetscentral/net/mimebuilder.cc
  *  \brief Networking: MIME builder
  *  \author Stefan Reuther
  *
  *  RFC 1867 Form-based file upload
  *  RFC 2046 MIME
  *  RFC 2184 MIME parameter value and encoded word extensions
  *  RFC 2388 multipart/form-data
  *
  *  application/x-www-form-urlencoded: name=value&name=value...
  *  (some reference this as application/x-url-encoded)
  *
  *  multipart/form-data:
  *      Content-type: multipart/form-data; boundary=AaB03x
  *
  *      --AaB03x
  *      content-disposition: form-data; name="field1"
  *
  *      Joe Blow
  *      --AaB03x
  *      content-disposition: form-data; name="pics"; filename="file1.txt"
  *      Content-Type: text/plain
  *
  *      ... contents of file1.txt ...
  *      --AaB03x--
  *  or
  *      --AaB03x
  *      content-disposition: form-data; name="field1"
  *      content-type: text/plain;charset=windows-1250
  *      content-transfer-encoding: quoted-printable
  *
  *      Joe owes =80100.
  *      --AaB03x
  *
  *  Although an example in RFC 2388 indicates that individual parts
  *  of a form may be encoded (e.g. QP in the above one), browsers
  *  known to me don't do it, and Planetsserver does not handle it.
  *  Hence we send it unencoded.
  */

#include <ostream>
#include <string.h>
#include "planetscentral/net/mimebuilder.h"
#include "planetscentral/net/encoding.h"
#include "planetscentral/dbclient/connection.h"

static const char CRLF[] = "\r\n";

/** Constructor.
    \param ctype Content type. Can be "multipart/mixed", "multipart/form-data", etc.
    Can be blank if you wish a non-multipart content type and want to add the
    header yourself. */
MimeBuilder::MimeBuilder(string_t ctype)
    : m_Content(),
      m_ContentType(ctype),
      m_Base64Accumulator(),
      m_PendingCRLF(false),
      m_State(StateHeader)
{
    if (!m_ContentType.empty()) {
        m_Content.push_back(Element(string_t(), ContentTypeElement));
    }
}

/** Destructor. */
MimeBuilder::~MimeBuilder()
{ }

/** Ensure we are in the body of the message. In case we're still in
    the header, adds the blank line that separates them. */
void
MimeBuilder::flushHeaderAndStartBody()
{
    if (m_State == StateHeader) {
        m_State = StateBody;
        m_Content.push_back(Element(string_t(CRLF), TextElement));
    }
}

/** Flush base64 buffer. In case we have pending base64 characters,
    write them out to the message. */
void
MimeBuilder::flushBase64()
{
    if (m_Base64Accumulator.size()) {
        m_Content.push_back(Element(encodeBase64(m_Base64Accumulator) + CRLF, TextElement));
        m_Base64Accumulator.clear();
    }
    if (m_PendingCRLF) {
        m_Content.push_back(Element(CRLF, TextElement));
        m_PendingCRLF = false;
    }
}

/** Add header to message. Currently does no postprocessing of the header data.
    \param name   Name of header
    \param value  Value of header */
void
MimeBuilder::addHeader(string_t name, string_t value)
{
    // we should make sure we're actually in header state
    m_Content.push_back(Element(name + ": " + value + CRLF, TextElement));
}

/** Add line to message. Does not modify the line. */
void
MimeBuilder::addLine(string_t line)
{
    flushHeaderAndStartBody();
    m_Content.push_back(Element(line + CRLF, TextElement));
}

/** Add base64 data.
    \param data [in] Data to add
    \param size [in] Size of data to add */
void
MimeBuilder::addBase64(const char* data, std::size_t size)
{
    // We want 76 characters per line, that is 19 quads corresponding
    // to 19 input triplets i.e. 57 input characters
    flushHeaderAndStartBody();
    while (size) {
        std::size_t room = 57 - m_Base64Accumulator.size();
        std::size_t now = size > room ? room : size;
        m_Base64Accumulator.append(data, now);
        if (m_Base64Accumulator.size() == 57)
            flushBase64();
        data += now;
        size -= now;
    }
}

/** Add quoted-printable line. Does qp encoding before adding the line
    to the message.
    \param line [in] The line to add. */
void
MimeBuilder::addLineQP(string_t line)
{
    // FIXME: we should split long lines
    flushHeaderAndStartBody();

    string_t::size_type pos = 0, n;
    while ((n = line.find('\n', pos)) != line.npos) {
        m_Content.push_back(Element(encodeQP(string_t(line, pos, n-pos)) + CRLF, TextElement));
        pos = n+1;
    }
    m_Content.push_back(Element(encodeQP(string_t(line, pos)) + CRLF, TextElement));
}

/** Add raw data.

    When sending raw data, we must be extra careful when choosing the
    boundary line, to not accidentally trigger a split in the middle
    of the file. finish() checks the beginnings of all TextElement's.
    We therefore split the binary data into Element's in such a way
    that possible conflict points start a new Element.

    RFC 2046 says a boundary is recognized only after CRLF. For
    robustness, we assume conflict points at every LF. This should be
    as robust network-wise as it can get. It could be made a little
    more memory-efficient by splitting only at "\n--" places. */
void
MimeBuilder::addRawData(const char* data, std::size_t size)
{
    flushHeaderAndStartBody();

    using namespace std;
    const char* p;
    while (size != 0 && (p = static_cast<const char*>(memchr(data, '\n', size))) != 0) {
        std::size_t length = p - data + 1;
        m_Content.push_back(Element(string_t(data, length), TextElement));
        data += length;
        size -= length;
    }
    if (size != 0) {
        m_Content.push_back(Element(string_t(data, size), TextElement));
    }
    m_PendingCRLF = true;
}

/** Add a boundary. After this call, you can either write a new part
    using addHeader() etc., or use finish(). */
void
MimeBuilder::addBoundary()
{
    flushHeaderAndStartBody();
    flushBase64();
    m_Content.push_back(Element(string_t(), BoundaryElement));
    m_State = StateHeader;
}

/** Complete the MIME entity. This figures out the boundary string to
    use, and fills it in at appropriate places. */
void
MimeBuilder::finish()
{
    // figure out a good boundary
    string_t boundary = "--000";
    while (1) {
        // is it already in use?
        bool ok = true;
        for (std::list<Element>::iterator i = m_Content.begin(); i != m_Content.end(); ++i) {
            // The additional size test is because Borland C++ doesn't like
            // compare(a,b,c) being called on a string shorter than a+b.
            if (i->m_Text.size() >= boundary.size() && i->m_Text.compare(0, boundary.size(), boundary) == 0) {
                ok = false;
                break;
            }
        }

        // nope, use it.
        if (ok)
            break;

        // yes. Guess a new one.
        string_t::size_type i = 2;
        while (i < boundary.size() && boundary[i] == '9') {
            boundary[i] = '0';
            ++i;
        }
        if (i < boundary.size())
            ++boundary[i];
        else
            boundary += '0';
    }

    // now fill in the content
    string_t* last_boundary = 0;
    for (std::list<Element>::iterator i = m_Content.begin(); i != m_Content.end(); ++i) {
        if (i->m_Type == ContentTypeElement) {
            i->m_Text = "Content-Type: " + m_ContentType + "; boundary=" + boundary.substr(2) + CRLF;
        } else if (i->m_Type == BoundaryElement) {
            i->m_Text = boundary + CRLF;
            last_boundary = &i->m_Text;
        }
    }
    if (last_boundary)
        *last_boundary = boundary + "--" + CRLF;
}

/** Compute total size of this entity. */
std::size_t
MimeBuilder::computeSize() const
{
    std::size_t sum = 0;
    for (std::list<Element>::const_iterator i = m_Content.begin(); i != m_Content.end(); ++i) {
        sum += i->m_Text.size();
    }
    return sum;
}

/** Write this entity onto a stream. */
void
MimeBuilder::write(std::ostream& out, bool escapeDots)
{
    for (std::list<Element>::const_iterator i = m_Content.begin(); i != m_Content.end(); ++i) {
        if (escapeDots && !i->m_Text.empty() && i->m_Text[0] == '.') {
            out << '.';
        }
        out << i->m_Text;
    }
}

/** Removes everything before the first boundary. Effectively, that's
    all the headers plus the blank line separating headers from body. */
string_t
MimeBuilder::removeInitialHeaders()
{
    string_t result;
    while (!m_Content.empty() && m_Content.front().m_Type != BoundaryElement) {
        result += m_Content.front().m_Text;
        m_Content.pop_front();
    }
    return result;
}
