/**
  *  \file planetscentral/talk/mailout.cc
  *  \brief Talk: Mail output
  *
  *  Mail output is produced for two purposes:
  *  - notification mails
  *  - NNTP messages
  *
  *  Those are subtly different. Whereas notifications preserve some
  *  more forum specifics, NNTP messages are converted into a format
  *  which a skilled NNTP user prefers. In particular, post references
  *  are turned into message-Id links, and forum references are turned
  *  into newsgroup names.
  */

#include "planetscentral/talk/mailout.h"
#include "planetscentral/talk/user.h"
#include "util/misc.h"
#include "planetscentral/talk/text.h"
#include "planetscentral/talk/message.h"
#include "planetscentral/talk/render.h"
#include "planetscentral/talk/topic.h"
#include "planetscentral/talk/url.h"
#include "planetscentral/talk/db.h"
#include "planetscentral/talk/forum.h"

namespace {
    class MailRenderer {
     public:
        MailRenderer(string_t& result, bool forNNTP, RenderState& state, DbClient& db);

        void renderUserName(const string_t& name);
        string_t formatMessageId(const string_t& name);
        bool renderThreadId(const string_t& name);
        bool renderGameLink(const string_t& text);
        bool renderUserLink(const string_t& text);
        bool renderForumLink(const string_t& text);
        void renderPlaintext(const string_t& text);
        void renderCode(const string_t& text);
        void renderAttribution(const string_t& text);
        void renderPG(TextNode* n);
        void renderChildrenPG(TextNode* n);

        void renderInline(TextNode* n);
        void renderChildrenInline(TextNode* n);

        MailRenderer& withPrefix(string_t pfx, string_t second);

        void renderWord(string_t part);
        void renderSpace();
        void flushLine();

     private:
        void emitLine(const string_t& theLine);

        string_t& result;
        RenderState& state;
        DbClient& db;
        const bool forNNTP;

        string_t line;
        string_t word;
        string_t prefix;
        string_t second_prefix;
    };
}

MailRenderer::MailRenderer(string_t& result, bool forNNTP, RenderState& state, DbClient& db)
    : result(result),
      state(state),
      db(db),
      forNNTP(forNNTP),
      line(),
      word(),
      prefix(),
      second_prefix()
{ }

/** Render a user name. Renders the user's real name or screen name.
    Used for rendering attributions. */
void
MailRenderer::renderUserName(const string_t& name)
{
    // Map user name to user Id
    string_t userId = User::getUserIdFromLogin(db, name);
    if (userId.empty()) {
        renderWord(name);
    } else {
        User u(db, userId);
        string_t realName = u.getRealName();
        if (realName.empty()) {
            renderWord(u.getScreenName());
        } else {
            renderWord(realName);
        }
    }
}

/** Render a message Id.
    For NNTP, render as the RFC message Id.
    Otherwise, render a <post:XXX> pseudo-link. */
string_t
MailRenderer::formatMessageId(const string_t& name)
{
    int32_t mid;
    if (!forNNTP || !stringValue(name, mid)) {
        return "<post:" + name + ">";
    } else {
        Message m(db, mid);
        if (!m.exists()) {
            return "<post:" + name + ">";
        } else {
            return "<" +m.getRfcMessageId() + ">";
        }
    }
}

/** Render a thread Id.
    For NNTP, render it as the RFC message Id of the first posting.
    If it has no message Id (or we're not rendering for NNTP),
    returns false, causing the caller to produce a <thread:XXX>
    pseudo-link. */
bool
MailRenderer::renderThreadId(const string_t& name)
{
    int32_t tid;
    if (!forNNTP || !stringValue(name, tid)) {
        return false;
    }

    Topic t(db, tid);
    if (!t.exists()) {
        return false;
    }

    Message m(db, t.firstPostingId().get());
    if (!m.exists()) {
        return false;
    }

    renderWord("<" + m.getRfcMessageId() + ">");
    return true;
}

/** Render a game link.
    Always renders the complete link. */
bool
MailRenderer::renderGameLink(const string_t& text)
{
    // Parse game Id
    int32_t gameId;
    if (!stringValue(text, gameId) || gameId <= 0) {
        return false;
    }

    // Access game, check permissions
    // xref host/game.cc, Game::hasPermission
    DbSubtree root(db, GAME_ROOT);
    if (!root.intSetKey("all").contains(gameId)) {
        return false;
    }

    DbSubtree game(root.subtree(gameId));
    const string_t gameState = game.stringKey("state").get();
    if (gameState != "joining" && gameState != "running" && gameState != "finished") {
        return false;
    }

    const string_t gameType = game.stringKey("type").get();
    if (gameType != "unlisted"
        && gameType != "public"
        && game.stringKey("owner").get() != state.getUser()
        && !game.hashKey("users").hasField(state.getUser()))
    {
        return false;
    }

    // OK, we are allowed to access it. Get its name.
    const string_t name = game.stringKey("name").get();
    renderWord("<" + state.getBaseUrl() + makeGameUrl(gameId, name) + ">");
    return true;
}

bool
MailRenderer::renderUserLink(const string_t& text)
{
    // Map user name to user Id
    string_t userId = User::getUserIdFromLogin(db, text);
    if (userId.empty()) {
        return false;
    }
    User u(db, userId);

    // Build it
    renderWord("<" + state.getBaseUrl() + USER_BASE_URL + u.getLoginName() + ">");
    return true;
}

bool
MailRenderer::renderForumLink(const string_t& text)
{
    // Parse forum Id
    int32_t forumId;
    if (!stringValue(text, forumId) || forumId <= 0) {
        return false;
    }

    // Access forum, check permissions
    Forum forum(db, forumId);
    if (!forum.exists()) {
        // FIXME permission check!
        return false;
    }

    // Can we render as newsgroup name?
    if (forNNTP) {
        string_t ngName = forum.getNewsgroup();
        if (!ngName.empty()) {
            renderWord("<news:" + ngName + ">");
            return true;
        }
    }

    // Render as link
    const string_t name = forum.name().get();
    renderWord("<" + state.getBaseUrl() + makeForumUrl(forumId, name) + ">");
    return true;
}

void
MailRenderer::renderPlaintext(const string_t& text)
{
    static const char WHITESPACE[] = " \t\r\n";
    string_t::size_type i = text.find_first_not_of(WHITESPACE);
    if (i != 0 && i != string_t::npos) {
        renderSpace();
    }
    while (i != string_t::npos) {
        string_t::size_type j = text.find_first_of(WHITESPACE, i);
        if (j == string_t::npos) {
            renderWord(string_t(text, i));
            break;
        } else {
            renderWord(string_t(text, i, j-i));
            renderSpace();
            i = text.find_first_not_of(WHITESPACE, j);
        }
    }
}

/** Render preformatted text. */
void
MailRenderer::renderCode(const string_t& text)
{
    std::size_t i = 0;
    std::size_t j;
    while ((j = text.find_first_of("\r\n", i)) != text.npos) {
        renderWord("  " + string_t(text, i, j-i));
        flushLine();
        i = j+1;
        if (text[j] == '\r' && i < text.size() && text[i] == '\n') {
            ++i;
        }
    }
    if (i != text.size()) {
        renderWord("  " + string_t(text, i, text.size()-i));
        flushLine();
    }
}

/** Render attribution of a message. */
void
MailRenderer::renderAttribution(const string_t& text)
{
    if (!text.empty()) {
        string_t::size_type i = text.rfind(';');
        if (i != text.npos && i != text.size()-1) {
            // It is the form "user;posting"
            renderWord("* ");
            renderUserName(string_t(text, 0, i));
            renderWord(" in ");
            renderWord(formatMessageId(string_t(text, i+1)));
            renderWord(":");
            flushLine();
        } else {
            // Just a user name
            renderWord("* ");
            renderUserName(text);
            renderWord(":");
            flushLine();
        }
    }
}

/** Render a paragraph/group. These always produce complete lines. */
void
MailRenderer::renderPG(TextNode* n)
{
    string_t::size_type x = result.size();
    switch (TextMajorKind(n->major)) {
     case maGroup:
        switch (TextGroupFormat(n->minor)) {
         case miGroupQuote:
            renderAttribution(n->text);
            if (prefix.empty()) {
                MailRenderer(*this).withPrefix("> ", "> ").renderChildrenPG(n);
            } else {
                MailRenderer(*this).withPrefix(">" + prefix, ">" + second_prefix).renderChildrenPG(n);
            }
            if (x != result.size()) {
                prefix = second_prefix;
            }
            break;
         case miGroupList:
            renderChildrenPG(n);
            break;
         case miGroupListItem:
            MailRenderer(*this).withPrefix(prefix + "* ", second_prefix + "  ").renderChildrenPG(n);
            if (x != result.size()) {
                prefix = second_prefix;
            }
            break;
         case miGroupRoot:
         default:
            renderChildrenPG(n);
            break;
        }
        break;

     case maParagraph:
        switch (TextParagraphFormat(n->minor)) {
         case miParCode:
            renderCode(n->rawTextContent());
            break;
         case miParCentered:    // centering is not rendered in mails
         case miParNormal:      // regular paragraphs
         default:               // default is regular paragraph
            renderChildrenInline(n);
            flushLine();
            break;
        }
     default:
        break;
    }
}

/** Render a paragraph/group container. */
void
MailRenderer::renderChildrenPG(TextNode* n)
{
    for (ptr_vector<TextNode>::size_type i = 0, e = n->children.size(); i != e; ++i) {
        if (i != 0) {
            result += prefix;
            result += "\n";
        }
        renderPG(n->children[i]);
    }
}

/** Render inline markup. Inline markup produced by accumulating and
    word-wrapping it in \c line. */
void
MailRenderer::renderInline(TextNode* n)
{
    switch (TextMajorKind(n->major)) {
     case maPlain:
        renderPlaintext(n->text);
        break;
     case maInline:
        renderChildrenInline(n);
        break;
     case maInlineAttr:
        renderChildrenInline(n);
        break;
     case maLink:
        renderChildrenInline(n);
        renderSpace();
        switch (TextLinkFormat(n->minor)) {
         case miLinkUrl:
            renderWord("<" + n->text + ">");
            break;
         case miLinkEmail:
            renderWord("<mailto:" + n->text + ">");
            break;
         case miLinkPost:
            renderWord(formatMessageId(n->text));
            break;
         case miLinkThread:
            if (!renderThreadId(n->text)) {
                renderWord("<thread:" + n->text + ">");
            }
            break;
         case miLinkGame:
            if (!renderGameLink(n->text)) {
                renderWord("<game:" + n->text + ">");
            }
            break;
         case miLinkUser:
            if (!renderUserLink(n->text)) {
                renderWord("<user:" + n->text + ">");
            }
            break;
         case miLinkForum:
            if (!renderForumLink(n->text)) {
                renderWord("<forum:" + n->text + ">");
            }
            break;
        }
        break;

     case maSpecial:
        switch (TextSpecialFormat(n->minor)) {
         case miSpecialBreak:
            flushLine();
            break;

         case miSpecialImage:
            renderChildrenInline(n);
            renderSpace();
            renderWord("<" + n->text + ">");
            break;

         case miSpecialSmiley:
            renderWord(":" + n->text + ":");
            break;
        }
        break;

     case maParagraph:
     case maGroup:
        break;
    }
}

/** Render a inline markup container. */
void
MailRenderer::renderChildrenInline(TextNode* n)
{
    for (ptr_vector<TextNode>::size_type i = 0, e = n->children.size(); i != e; ++i) {
        renderInline(n->children[i]);
    }
}

MailRenderer&
MailRenderer::withPrefix(string_t pfx, string_t second)
{
    prefix = pfx;
    second_prefix = second;
    return *this;
}

void
MailRenderer::renderWord(string_t part)
{
    word += part;
}

void
MailRenderer::renderSpace()
{
    if (!word.empty()) {
        // Get permitted line size
        string_t::size_type lineLength = 72;
        if (prefix.size() > 30) {
            lineLength = prefix.size() + 42;
        }

        // Format it
        if (!line.empty() && prefix.size() + line.size() + word.size() >= lineLength) {
            emitLine(line);
            line.clear();
        }
        if (!line.empty()) {
            line += " ";
        }
        line += word;
        word.clear();
    }
}

void
MailRenderer::flushLine()
{
    renderSpace();
    if (!line.empty()) {
        emitLine(line);
        line.clear();
    }
}

void
MailRenderer::emitLine(const string_t& theLine)
{
    result += prefix;
    result += theLine;
    result += "\n";
    prefix = second_prefix;
}

string_t
renderMail(TextNode* n, bool forNNTP, RenderState& state, DbClient& db)
{
    string_t result;
    MailRenderer(result, forNNTP, state, db).renderChildrenPG(n);
    return result;
}
