/**
  *  \file planetscentral/talk/syntax.h
  *  \brief Talk: Syntax Highlighter, Base Class
  */
#ifndef C2SERVER_PLANETSCENTRAL_TALK_SYNTAX_H
#define C2SERVER_PLANETSCENTRAL_TALK_SYNTAX_H

#include "cpluslib/string.h"

/** Syntax highlighter, base class.
    To highlight some piece of text,
    - call init()
    - call scan() until it returns false

    Each scan() produces one segment of text with highlighting information.

    Derived classes override init() and scan().
    scan() can use Segment's methods to build the segment,
    either set(), or start()+finish().
    It can also assign the members manually. */
class SyntaxHighlighter {
 public:
    enum Format {
        DefaultFormat,          // Nothing in particular
        KeywordFormat,          // A keyword (C++: 'else')
        NameFormat,             // A name ('i' in 'int i', 'Dim i'; 'NumMinefields' in 'NumMinefields=1')
        StringFormat,           // A string
        CommentFormat,          // A comment
        Comment2Format,         // A special comment, e.g. "### foo" in ini files
        SectionFormat,          // A section divider, e.g. "%PHOST"
        ErrorFormat             // Unparseable
    };
    struct Segment {
        Format fmt;
        string_t::size_type start_pos;
        string_t::size_type length;
        string_t link;
        string_t info;
        void set(Format fmt, string_t::size_type start, string_t::size_type length);
        void start(string_t::size_type start);
        void finish(Format fmt, string_t::size_type end);
    };

    /** Virtual destructor. */
    virtual ~SyntaxHighlighter();

    /** Initialize. Sets the string to parse and reset state.
        The string is not copied. */
    virtual void init(const string_t& str) = 0;

    /** Extract a segment.
        \param result [out] Space for result
        \return true if a result was produced, false if the end was reached */
    virtual bool scan(Segment& result) = 0;
};

/** A simple syntax highlighter that does nothing. */
class NullSyntaxHighlighter : public SyntaxHighlighter {
 public:
    NullSyntaxHighlighter();
    virtual ~NullSyntaxHighlighter();
    virtual void init(const string_t& str);
    virtual bool scan(Segment& result);

 private:
    string_t::size_type pos;
    string_t::size_type total;
};

#endif
